from copy import deepcopy
from typing import List

import networkx

from metrics_layer.core.model.base import MetricsLayerBase
from metrics_layer.core.model.definitions import Definitions
from metrics_layer.core.sql.query_errors import ParseError


class MetricsLayerDesign:
    """ """

    def __init__(self, no_group_by: bool, query_type: str, field_lookup: dict, explore, project) -> None:
        self.no_group_by = no_group_by
        self.query_type = query_type
        self.field_lookup = field_lookup
        self.explore = explore
        self.project = project
        self._joins = None

    def views(self) -> List[MetricsLayerBase]:
        return self.project.views(explore_name=self.explore.name)

    def joins(self) -> List[MetricsLayerBase]:
        if self._joins is None:
            _, access_filter_fields = self.get_access_filter()
            fields_in_query = list(self.field_lookup.values()) + access_filter_fields
            required_views = list(set([v for field in fields_in_query for v in field.required_views()]))
            if self.explore.always_join:
                required_views.extend(self.explore.always_join)

            joins_needed_for_query = []
            for view_name in reversed(sorted(required_views)):
                joins_needed_for_query.extend(self._find_needed_joins(view_name, joins_needed_for_query))
            self._joins = self._sort_joins(joins_needed_for_query)
        return self._joins

    def _find_needed_joins(self, view_name: str, joins_already_added: list):
        joins_to_add = []

        join_already_added = any(view_name == j.from_ for j in joins_already_added)
        if not join_already_added and view_name != self.explore.from_:
            join = self.explore.get_join(view_name, by_view_name=True)
            if join is None:
                raise ValueError(
                    f"Could not locate join from view {view_name} for explore {self.explore.name}"
                )
            joins_to_add.append(join)
            for view_name in join.required_views():
                joins_to_add.extend(self._find_needed_joins(view_name, joins_already_added + [join]))
        return joins_to_add

    def _sort_joins(self, joins_needed: list):
        if len(joins_needed) == 0:
            return []

        self._join_graph = networkx.DiGraph()
        for join in joins_needed:
            for view_name in join.required_views():
                if view_name != join.from_:
                    self._join_graph.add_edge(view_name, join.from_, relationship=join.relationship)

        self._ordered_join_names = list(networkx.topological_sort(self._join_graph))
        # Skip the first one because that's *always* the base of the explore
        return [self.explore.get_join(name, by_view_name=True) for name in self._ordered_join_names[1:]]

    @staticmethod
    def chain_decomposition(graph, root):
        result = []
        for decomp in networkx.chain_decomposition(graph, root=root):
            if decomp[0][0] == root:
                result = []
                for edge in decomp:
                    result.append(edge[0])
        return result

    def functional_pk(self):
        sorted_joins = self.joins()
        if len(sorted_joins) == 0:
            return self.get_view(self.base_view_name).primary_key
        elif any(j.relationship == "many_to_many" for j in sorted_joins):
            # There is no functional primary key if there is a many_to_many join
            return Definitions.does_not_exist
        elif all(j.relationship in {"many_to_one", "one_to_one"} for j in sorted_joins):
            # The functional primary key is the key to the base join is all joins are many_to_one
            return self.get_view(self.base_view_name).primary_key
        else:
            base_view = self.get_view(self.base_view_name)
            primary_key_view_name = self._derive_primary_key_view(base_view, sorted_joins)

            if primary_key_view_name == Definitions.does_not_exist:
                return Definitions.does_not_exist
            elif primary_key_view_name != base_view.name:
                primary_key_view = self.get_view(primary_key_view_name)
                return primary_key_view.primary_key
            return base_view.primary_key

    def _derive_primary_key_view(self, base_view, sorted_joins: list):
        # if the branch is from the base and many_to_one the base is the same
        # if the branch is from a many_to_one to the base and many_to_one or one_to_one it's the same
        # if the branch is from a many_to_one to the base and one_to_many it's now many_to_many

        # if the branch is from the base and one_to_one the base is the same
        # if the branch is from a one_to_one to the base and many_to_one or one_to_one it's the same
        # if the branch is from a one_to_one to the base and one_to_many the base is the new one

        # if the branch is from the base and one_to_many the base is the new one
        # if the branch is from a one_to_many to the base and many_to_one or one_to_one it's
        #   the one referenced in the one_to_many
        # if the branch is from a one_to_many to the base and one_to_many the base is now
        #   the newest one_to_many ref

        working_base = base_view.name
        previous_join_type = None
        resolve_sequences = []
        for j in sorted_joins:
            if j.relationship == "many_to_many":
                return Definitions.does_not_exist

            # Since there can be many paths independent of each other, I have to check each
            # path to find what should be the base and comapre the results, if one path claims
            # to have a different base from the initial one then that path must the the only
            # one to make that claim (e.g. 2 one_to_many joins going from the same base
            # is a many_to_many resulting pk). This logic is in the
            path = networkx.shortest_path(self._join_graph, working_base, j.from_)

            base_sequence = deepcopy([base_view.name])
            previous_join_type = None
            for from_, to_ in enumerate(range(1, len(path))):
                relationship = self._join_graph[path[from_]][path[to_]]["relationship"]
                if relationship == "one_to_many" and previous_join_type == "many_to_one":
                    return Definitions.does_not_exist
                elif relationship == "one_to_many":
                    base_sequence.append(j.from_)
                previous_join_type = relationship
            resolve_sequences.append(base_sequence)
        primary_key = self._pk_from_join_sequences(resolve_sequences)
        return primary_key

    def _pk_from_join_sequences(self, join_sequences: list):

        lengths, final_selections = [], []
        for sequence in join_sequences:
            lengths.append(len(sequence))
            final = sequence[-1]
            final_selections.append(final)

        # If all conclusions are the same than that's the right pk
        if len(set(final_selections)) == 1:
            return final_selections[0]

        # If there is disagreement in the final conclusions, we need to check for sub lists
        # E.g. if these are the join sequences:
        # [customers]
        # [customers, orders]
        # [customers, orders, order_lines]
        # The above is a pk of order_lines, because the differing conclusions are just sub oaths

        # e.g in this case they are not two sub-paths but actually different join paths
        # [customers, orders]
        # [customers, discounts]
        # The above is many_to_many
        longest_idx = lengths.index(max(lengths))
        longest_sequence = join_sequences[longest_idx]
        longest_final = final_selections[longest_idx]
        for sequence in join_sequences:
            if sequence[-1] != longest_final and not self._is_sublist(longest_sequence, sequence):
                return Definitions.does_not_exist

        return longest_final

    @staticmethod
    def _is_sublist(main_list: list, sublist: list):
        n_contained_lists = len(main_list) - len(sublist) + 1
        return any(main_list[idx : idx + len(sublist)] == sublist for idx in range(n_contained_lists))

    def get_view(self, name: str) -> MetricsLayerBase:
        try:
            return next(t for t in self.views() if t.name == name)
        except StopIteration:
            raise ParseError(f"View {name} not found in explore {self.explore.name}")

    def get_join(self, name: str) -> MetricsLayerBase:
        return next((j for j in self.joins() if j.name == name), None)

    def get_field(self, field_name: str) -> MetricsLayerBase:
        return self.project.get_field(field_name, explore_name=self.explore.name)

    def get_access_filter(self):
        if self.explore.access_filters:
            conditions, fields = [], []
            for condition_set in self.explore.access_filters:
                field = self.project.get_field(condition_set["field"], explore_name=self.explore.name)
                sql = field.sql_query(self.query_type)
                user_attribute_value = condition_set["user_attribute"]

                if self.project._user and self.project._user.get(user_attribute_value):
                    condition = f"{sql} = '{self.project._user[user_attribute_value]}'"
                    conditions.append(condition)
                    fields.append(field)
            return " and ".join(conditions), fields
        return None, []

    @property
    def base_view_name(self):
        return self.explore.from_



import sqlparse
from sqlparse.tokens import Name, Punctuation

from metrics_layer.core.parse.config import ConfigError, MetricsLayerConfiguration
from metrics_layer.core.sql.query_design import MetricsLayerDesign
from metrics_layer.core.sql.query_generator import MetricsLayerQuery


class SingleSQLQueryResolver:
    """
    Method of resolving the explore name:
        if there is not explore passed (using the format explore_name.field_name), we'll search for
        just the field name and iff that field is used in only one explore, set that as the active explore.
            - Any fields specified that are not in that explore will raise an error

        if it's passed explicitly, use the first metric's explore, and raise an error if anything conflicts
        with that explore
    """

    def __init__(
        self,
        metrics: list,
        dimensions: list = [],
        where: str = None,  # Either a list of json or a string
        having: str = None,  # Either a list of json or a string
        order_by: str = None,  # Either a list of json or a string
        config: MetricsLayerConfiguration = None,
        **kwargs,
    ):
        self.field_lookup = {}
        self.no_group_by = False
        self.verbose = kwargs.get("verbose", False)
        self.select_raw_sql = kwargs.get("select_raw_sql", [])
        self.explore_name = kwargs.get("explore_name")
        self.suppress_warnings = kwargs.get("suppress_warnings", False)
        self.limit = kwargs.get("limit")
        self.return_pypika_query = kwargs.get("return_pypika_query")
        self.config = config
        self.project = self.config.project
        self.metrics = metrics
        self.dimensions = dimensions
        self.parse_field_names(where, having, order_by)

        if not self.explore_name:
            self.explore_name = self.derive_explore(self.verbose)

        self.explore = self.project.get_explore(self.explore_name)
        try:
            self.connection = self.config.get_connection(self.explore.model.connection)
        except ConfigError:
            self.connection = None

        if "query_type" in kwargs:
            self.query_type = kwargs["query_type"]
        elif self.connection:
            self.query_type = self.connection.type
        else:
            raise ConfigError(
                "Could not determine query_type. Please have connection information for "
                "your warehouse in the configuration or explicitly pass the "
                "'query_type' argument to this function"
            )
        self.parse_input()

    def get_query(self, semicolon: bool = True):
        self.design = MetricsLayerDesign(
            no_group_by=self.no_group_by,
            query_type=self.query_type,
            field_lookup=self.field_lookup,
            explore=self.explore,
            project=self.project,
        )

        query_definition = {
            "metrics": self.metrics,
            "dimensions": self.dimensions,
            "where": self.where,
            "having": self.having,
            "order_by": self.order_by,
            "select_raw_sql": self.select_raw_sql,
            "limit": self.limit,
            "return_pypika_query": self.return_pypika_query,
        }
        query = MetricsLayerQuery(
            query_definition, design=self.design, suppress_warnings=self.suppress_warnings
        ).get_query(semicolon=semicolon)

        return query

    def derive_explore(self, verbose: bool):
        # Only checking metrics when they exist reduces the number of obvious explores a user has to specify
        if len(self.metrics) > 0:
            all_fields = self.metrics
        else:
            all_fields = self.dimensions

        if len(all_fields) == 0:
            raise ValueError("You need to include at least one metric or dimension for the query to run")

        initial_field = all_fields[0]
        working_explore_name = self.project.get_explore_from_field(initial_field)
        for field_name in all_fields[1:]:
            explore_name = self.project.get_explore_from_field(field_name)
            if explore_name != working_explore_name:
                raise ValueError(
                    f"""The explore found in metric {initial_field}, {working_explore_name}
                    does not match the explore found in {field_name}, {explore_name}"""
                )

        if verbose:
            print(f"Setting query explore to: {working_explore_name}")
        return working_explore_name

    def parse_input(self):
        # TODO handle this case in the future
        if self.explore.symmetric_aggregates == "no":
            raise NotImplementedError(
                "MetricsLayer does not currently support turning off symmetric aggregates"
            )

        all_field_names = self.metrics + self.dimensions
        if len(set(all_field_names)) != len(all_field_names):
            # TODO improve this error message
            raise ValueError("Ambiguous field names in the metrics and dimensions")

        for name in self.metrics:
            self.field_lookup[name] = self.get_field_with_error_handling(name, "Metric")

        # Dimensions exceptions:
        #   They are coming from a different explore than the metric, not joinable (handled in get_field)
        #   They are not found in the selected explore (handled here)

        for name in self.dimensions:
            field = self.get_field_with_error_handling(name, "Dimension")
            # We will not use a group by if the primary key of the main resulting table is included
            if field.primary_key == "yes" and field.view.name == self.explore.from_:
                self.no_group_by = True
            self.field_lookup[name] = field

        for name in self._where_field_names:
            self.field_lookup[name] = self.get_field_with_error_handling(name, "Where clause field")

        for name in self._having_field_names:
            self.field_lookup[name] = self.get_field_with_error_handling(name, "Having clause field")

        for name in self._order_by_field_names:
            self.field_lookup[name] = self.get_field_with_error_handling(name, "Order by field")

    def get_field_with_error_handling(self, field_name: str, error_prefix: str):
        field = self.project.get_field(field_name, explore_name=self.explore_name)
        if field is None:
            raise ValueError(f"{error_prefix} {field_name} not found in explore {self.explore_name}")
        return field

    def parse_field_names(self, where, having, order_by):
        self.where = self._check_for_dict(where)
        if self._is_literal(self.where):
            self._where_field_names = self.parse_identifiers_from_clause(self.where)
        else:
            self._where_field_names = self.parse_identifiers_from_dicts(self.where)

        self.having = self._check_for_dict(having)
        if self._is_literal(self.having):
            self._having_field_names = self.parse_identifiers_from_clause(self.having)
        else:
            self._having_field_names = self.parse_identifiers_from_dicts(self.having)

        self.order_by = self._check_for_dict(order_by)
        if self._is_literal(self.order_by):
            self._order_by_field_names = self.parse_identifiers_from_clause(self.order_by)
        else:
            self._order_by_field_names = self.parse_identifiers_from_dicts(self.order_by)

    @staticmethod
    def _is_literal(clause):
        return isinstance(clause, str) or clause is None

    @staticmethod
    def parse_identifiers_from_clause(clause: str):
        if clause is None:
            return []
        generator = list(sqlparse.parse(clause)[0].flatten())

        field_names = []
        for i, token in enumerate(generator):
            not_already_added = i == 0 or str(generator[i - 1]) != "."
            if token.ttype == Name and not_already_added:
                field_names.append(str(token))

            if token.ttype == Punctuation and str(token) == ".":
                if generator[i - 1].ttype == Name and generator[i + 1].ttype == Name:
                    field_names[-1] += f".{str(generator[i+1])}"
        return field_names

    @staticmethod
    def parse_identifiers_from_dicts(conditions: list):
        try:
            return [cond["field"] for cond in conditions]
        except KeyError:
            for cond in conditions:
                if "field" not in cond:
                    break
            raise KeyError(f"Identifier was missing required 'field' key: {cond}")

    @staticmethod
    def _check_for_dict(conditions: list):
        if isinstance(conditions, dict):
            return [conditions]
        return conditions



import functools
import hashlib
import json
from collections import Counter

from .base import AccessDeniedOrDoesNotExistException
from .dashboard import Dashboard
from .explore import Explore
from .field import Field
from .model import AccessGrant, Model
from .view import View


class Project:
    """
    Higher level abstraction for the whole project
    """

    def __init__(
        self,
        models: list,
        views: list,
        dashboards: list = [],
        looker_env: str = None,
        connection_lookup: dict = {},
        manifest=None,
    ):
        self._models = models
        self._views = views
        self._dashboards = dashboards
        self.looker_env = looker_env
        self.connection_lookup = connection_lookup
        self.manifest = manifest
        self.manifest_exists = manifest and manifest.exists()
        self._user = None

    def __repr__(self):
        text = "models" if len(self._models) != 1 else "model"
        return f"<Project {len(self._models)} {text} user={self._user}>"

    def __hash__(self):
        model_str = json.dumps(self._models, sort_keys=True)
        view_str = json.dumps(self._views, sort_keys=True)
        dash_str = json.dumps(self._dashboards, sort_keys=True)
        conn_str = json.dumps(self.connection_lookup, sort_keys=True)
        user_str = "" if not self._user else json.dumps(self._user, sort_keys=True)
        string_to_hash = model_str + view_str + dash_str + conn_str + user_str + str(self.looker_env)
        result = hashlib.md5(string_to_hash.encode("utf-8"))
        return int(result.hexdigest(), base=16)

    def set_user(self, user: dict):
        self._user = user

    def validate(self):
        all_errors = []
        for model in self.models():
            all_errors.extend(model.collect_errors())

        for explore in self.explores():
            errors = explore.validate_fields()
            all_errors.extend(errors)

        for dashboard in self.dashboards():
            errors = dashboard.collect_errors()
            all_errors.extend(errors)

        all_errors.extend(self._validate_dashboard_names())

        return list(sorted(set(all_errors), key=lambda x: all_errors.index(x)))

    def _validate_dashboard_names(self):
        # We need to make sure the unique identifiers for the dashboards are actually unique
        errors = []
        dashboard_names = [d.name for d in self.dashboards()]
        name_frequency = Counter(dashboard_names).most_common()
        for name, frequency in name_frequency:
            if frequency > 1:
                msg = f"Dashboard name {name} appears {frequency} times, make sure dashboard names are unique"
                errors.append(msg)
            else:
                break
        return errors

    def _all_dashboards(self):
        dashboards = []
        for d in self._dashboards:
            dashboard = Dashboard(d, project=self)
            user_allowed = self.can_access_dashboard(dashboard)
            if user_allowed:
                dashboards.append(dashboard)
        return dashboards

    def dashboards(self) -> list:
        return self._all_dashboards()

    def get_dashboard(self, dashboard_name: str) -> Model:
        try:
            return next((d for d in self.dashboards() if d.name == dashboard_name))
        except StopIteration:
            raise AccessDeniedOrDoesNotExistException(
                f"Could not find or you do not have access to dashboard {dashboard_name}",
                object_name=dashboard_name,
                object_type="dashboard",
            )

    def models(self) -> list:
        return [Model(m) for m in self._models]

    def get_model(self, model_name: str) -> Model:
        return next((m for m in self.models() if m.name == model_name), None)

    def access_grants(self):
        return [AccessGrant(g) for m in self.models() for g in m.access_grants]

    def get_access_grant(self, grant_name: str):
        return next((ag for ag in self.access_grants() if ag.name == grant_name), None)

    def can_access_dashboard(self, dashboard: Dashboard):
        return self._can_access_object(dashboard)

    def can_access_explore(self, explore: Explore):
        return self._can_access_object(explore)

    def can_access_join(self, join, explore: Explore):
        can_access_explore = self.can_access_explore(explore)
        return self._can_access_object(join) and can_access_explore

    def can_access_view(self, view: View):
        return self._can_access_object(view)

    def can_access_field(self, field):
        can_access_view = self.can_access_view(field.view)
        return self._can_access_object(field) and can_access_view

    def can_access_merged_field(self, field, explore: Explore):
        can_access_explore = self.can_access_explore(explore)
        return self._can_access_object(field) and can_access_explore

    def _can_access_object(self, obj):
        if self._user is not None:
            if obj.required_access_grants:
                decisions = []
                for grant_name in obj.required_access_grants:
                    grant = self.get_access_grant(grant_name)
                    user_attribute_value = self._user.get(grant.user_attribute)

                    if user_attribute_value is None:
                        decision = True
                    else:
                        decision = user_attribute_value in grant.allowed_values
                    decisions.append(decision)

                # We use all here because the condition between access conditions is AND
                return all(decisions)
        return True

    def _all_explores(self):
        explores = []
        for m in self.models():
            for e in m.explores:
                explore = Explore({**e, "model": m}, project=self)
                user_allowed = self.can_access_explore(explore)
                if user_allowed:
                    explores.append(explore)
        return explores

    def explores(self, show_hidden: bool = True) -> list:
        explores = self._all_explores()
        if show_hidden:
            return explores
        return [e for e in explores if e.hidden == "no" or not e.hidden]

    def get_explore(self, explore_name: str) -> Explore:
        try:
            return next((e for e in self.explores() if e.name == explore_name))
        except StopIteration:
            raise AccessDeniedOrDoesNotExistException(
                f"Could not find or you do not have access to explore {explore_name}",
                object_name=explore_name,
                object_type="explore",
            )

    def _all_views(self, explore):
        views = []
        for v in self._views:
            view = View({**v, "explore": explore}, project=self)
            if self.can_access_view(view):
                views.append(view)
        return views

    def views(self, explore_name: str = None, explore: Explore = None) -> list:
        if explore_name:
            return self.views_with_explore(explore_name=explore_name)
        return self._all_views(explore)

    def views_with_explore(self, explore_name: str = None, explore: Explore = None):
        if explore_name and explore is None:
            explore = self.get_explore(explore_name)
        view_names = [explore.from_] + [j.from_ for j in explore.joins()]
        return [v for v in self.views(explore=explore) if v.name in view_names]

    def get_view(self, view_name: str, explore: Explore = None) -> View:
        try:
            return next((v for v in self.views(explore=explore) if v.name == view_name))
        except StopIteration:
            raise AccessDeniedOrDoesNotExistException(
                f"Could not find or you do not have access to view {view_name}",
                object_name=view_name,
                object_type="view",
            )

    def sets(self, view_name: str = None, explore_name: str = None):
        if explore_name:
            views = self.views(explore_name=explore_name)
        elif view_name:
            try:
                views = [self.get_view(view_name)]
            except AccessDeniedOrDoesNotExistException:
                views = []
        else:
            views = self.views()

        all_sets = []
        for view in views:
            all_sets.extend(view.list_sets())
        return all_sets

    def get_set(self, set_name: str, view_name: str = None):
        if view_name:
            sets = self.sets(view_name=view_name)
        else:
            sets = self.sets()
        return next((s for s in sets if s.name == set_name), None)

    @functools.lru_cache(maxsize=None)
    def fields(
        self,
        explore_name: str = None,
        view_name: str = None,
        show_hidden: bool = True,
        expand_dimension_groups: bool = False,
        show_excluded: bool = False,
    ) -> list:
        if explore_name is None and view_name is None:
            return self._all_fields(show_hidden, expand_dimension_groups)
        elif view_name and explore_name:
            fields = self._explore_fields(explore_name, show_hidden, expand_dimension_groups, show_excluded)
            return [f for f in fields if f.view.name == view_name]
        elif view_name:
            return self._view_fields(view_name, show_hidden, expand_dimension_groups)
        else:
            return self._explore_fields(explore_name, show_hidden, expand_dimension_groups, show_excluded)

    def _all_fields(self, show_hidden: bool, expand_dimension_groups: bool):
        return [f for v in self.views() for f in v.fields(show_hidden, expand_dimension_groups)]

    def _view_fields(
        self,
        view_name: str,
        show_hidden: bool = True,
        expand_dimension_groups: bool = False,
        explore: Explore = None,
    ):
        view = self.get_view(view_name, explore=explore)
        if not view:
            plus_explore = f" in explore {explore.name}" if explore else ""
            raise ValueError(f"Could not find a view matching the name {view_name}{plus_explore}")
        return view.fields(show_hidden, expand_dimension_groups)

    def _explore_fields(
        self, explore_name: str, show_hidden: bool, expand_dimension_groups: bool, show_excluded: bool
    ):
        explore = self.get_explore(explore_name)
        if show_excluded:
            valid_views = self.views_with_explore(explore=explore)
            return [f for v in valid_views for f in v.fields(show_hidden, expand_dimension_groups)]
        return explore.explore_fields(show_hidden, expand_dimension_groups, show_excluded)

    @functools.lru_cache(maxsize=None)
    def get_field(
        self, field_name: str, explore_name: str = None, view_name: str = None, show_excluded: bool = False
    ) -> Field:
        # Handle the case where the explore syntax is passed: explore_name.view_name.field_name
        if "." in field_name:
            specified_explore_name, specified_view_name, field_name = Field.field_name_parts(field_name)
            if view_name and specified_view_name != view_name:
                raise ValueError(
                    f"You specificed two different view names {specified_view_name} and {view_name}"
                )
            if specified_explore_name and explore_name and specified_explore_name != explore_name:
                raise ValueError(
                    f"You specificed two different explore names {specified_explore_name} and {explore_name}"
                )
            view_name = specified_view_name
            if specified_explore_name:
                explore_name = specified_explore_name

        field_name = field_name.lower()

        fields = self.fields(
            explore_name=explore_name,
            view_name=view_name,
            expand_dimension_groups=True,
            show_excluded=show_excluded,
        )
        matching_fields = [f for f in fields if f.equal(field_name)]
        return self._matching_field_handler(matching_fields, field_name, explore_name, view_name)

    def get_explore_from_field(self, field_name: str):
        # If it's specified this is really easy
        explore_name, view_name, to_match = Field.field_name_parts(field_name)
        if explore_name is not None:
            return explore_name

        # If it's not we have to check all explores to make sure the field isn't ambiguously referenced
        all_fields_with_explore_duplicates = []
        for explore in self.explores():
            for view in self.views_with_explore(explore_name=explore.name):
                all_fields_with_explore_duplicates.extend(view.fields(expand_dimension_groups=True))

        matching_fields = [f for f in all_fields_with_explore_duplicates if f.alias() == to_match]
        if view_name:
            matching_fields = [f for f in matching_fields if f.view.name == view_name]
        match = self._matching_field_handler(matching_fields, field_name)
        return match.view.explore.name

    def _matching_field_handler(
        self, matching_fields: list, field_name: str, explore_name: str = None, view_name: str = None
    ):
        if len(matching_fields) == 1:
            return matching_fields[0]

        elif len(matching_fields) > 1:
            # if exerything is in the same view we can pick the explore that uses this view as the from_ view
            if len(set([f.view.name for f in matching_fields])) == 1:
                view_name = matching_fields[0].view.name
                matching_explores = [e for e in self.explores() if e.from_ == view_name]
                # This method will work iff there is exactly one matchng explore
                if len(matching_explores) == 1:
                    for f in matching_fields:
                        if f.view.explore.name == matching_explores[0].name:
                            return f

            matching_names = [f.id() for f in matching_fields]
            explore_text = f", in explore {explore_name}" if explore_name else ""
            view_text = f", in view {view_name}" if view_name else ""
            err_msg = (
                f"Multiple fields found for the name {field_name}{explore_text}{view_text}"
                f" - those fields were {matching_names} \n\nPlease specify a "
                "view name like this: 'view_name.field_name' or "
                "an explore and view like this 'explore_name.view_name.field_name'"
                "\n\nor pass the argument 'explore_name' to the function, to set the explore"
            )
            raise ValueError(err_msg)
        elif field_name == "count" and (explore_name or view_name):
            definition = {"type": "count", "name": "count", "field_type": "measure"}
            if explore_name and not view_name:
                view_name = self.get_explore(explore_name).from_
            return Field(definition, view=self.get_view(view_name))
        else:
            err_msg = f"Field {field_name} not found"
            if explore_name:
                err_msg += f" in explore {explore_name}"
            if view_name:
                err_msg += f" in view {view_name}"
            err_msg += ", please check that this field exists AND that you have access to it. \n\n"
            err_msg += "If this is a dimension group specify the group parameter, if not already specified, "
            err_msg += "for example, with a dimension group named 'order' with timeframes: [raw, date, month]"
            err_msg += " specify 'order_raw' or 'order_date' or 'order_month'"
            raise AccessDeniedOrDoesNotExistException(err_msg, object_name=field_name, object_type="field")

    def resolve_dbt_ref(self, ref_name: str, view_name: str = None):
        if not self.manifest_exists:
            raise ValueError(
                f"Could not find a dbt project co-located with this "
                f"project to resolve the dbt ref('{ref_name}') in view {view_name}"
            )
        return self.manifest.resolve_name(ref_name)



from collections import defaultdict
from copy import deepcopy

from metrics_layer.core.parse.config import MetricsLayerConfiguration
from metrics_layer.core.sql.query_merged_results import MetricsLayerMergedResultsQuery
from metrics_layer.core.sql.single_query_resolve import SingleSQLQueryResolver


class SQLQueryResolver(SingleSQLQueryResolver):
    """
    Method of resolving the explore name:
        if there is not explore passed (using the format explore_name.field_name), we'll search for
        just the field name and iff that field is used in only one explore, set that as the active explore.
            - Any fields specified that are not in that explore will raise an error

        if it's passed explicitly, use the first metric's explore, and raise an error if anything conflicts
        with that explore
    """

    def __init__(
        self,
        metrics: list,
        dimensions: list = [],
        where: str = None,  # Either a list of json or a string
        having: str = None,  # Either a list of json or a string
        order_by: str = None,  # Either a list of json or a string
        config: MetricsLayerConfiguration = None,
        **kwargs,
    ):
        self.field_lookup = {}
        self.no_group_by = False
        self.verbose = kwargs.get("verbose", False)
        self.merged_result = kwargs.get("merged_result", False)
        self.select_raw_sql = kwargs.get("select_raw_sql", [])
        self.explore_name = kwargs.get("explore_name")
        self.suppress_warnings = kwargs.get("suppress_warnings", False)
        self.limit = kwargs.get("limit")
        self.config = config
        self.project = self.config.project
        self.metrics = metrics
        self.dimensions = dimensions
        self.where = where
        self.having = having
        self.order_by = order_by
        self.kwargs = kwargs
        self.connection = None

    def get_query(self, semicolon: bool = True):
        if self.merged_result:
            return self._get_merged_result_query(semicolon=semicolon)
        return self._get_single_query(semicolon=semicolon)

    def _get_single_query(self, semicolon: bool):
        resolver = SingleSQLQueryResolver(
            metrics=self.metrics,
            dimensions=self.dimensions,
            where=self.where,
            having=self.having,
            order_by=self.order_by,
            config=self.config,
            **self.kwargs,
        )
        query = resolver.get_query(semicolon)
        self.connection = resolver.connection
        return query

    def _get_merged_result_query(self, semicolon: bool):
        self.kwargs.pop("explore_name", None)

        self.parse_field_names(self.where, self.having, self.order_by)
        self.derive_sub_queries()

        explore_queries = {}
        for explore_name in self.explore_metrics.keys():
            metrics = [f.id(view_only=True) for f in self.explore_metrics[explore_name]]
            dimensions = [f.id(view_only=True) for f in self.explore_dimensions[explore_name]]

            # Overwrite the limit arg because these are subqueries
            kws = {**self.kwargs, "limit": None, "return_pypika_query": True}
            resolver = SingleSQLQueryResolver(
                metrics=metrics,
                dimensions=dimensions,
                where=self.explore_where[explore_name],
                having=[],
                order_by=[],
                config=self.config,
                explore_name=explore_name,
                **kws,
            )
            query = resolver.get_query(semicolon=False)
            explore_queries[explore_name] = query

        query_config = {
            "merged_metrics": self.merged_metrics,
            "explore_metrics": self.explore_metrics,
            "explore_dimensions": self.explore_dimensions,
            "having": self.having,
            "explore_queries": explore_queries,
            "explore_names": list(sorted(self.explore_metrics.keys())),
            "query_type": resolver.query_type,
            "limit": self.limit,
            "project": self.project,
        }
        merged_result_query = MetricsLayerMergedResultsQuery(query_config)
        query = merged_result_query.get_query(semicolon=semicolon)

        self.connection = resolver.connection
        return query

    def derive_sub_queries(self):
        # The different explores used are determined by the metrics referenced
        self.explore_metrics = defaultdict(list)
        self.merged_metrics = []
        for metric in self.metrics:
            explore_name = self.project.get_explore_from_field(metric)
            field = self.project.get_field(metric, explore_name=explore_name)
            if field.is_merged_result:
                self.merged_metrics.append(field)
            else:
                if field.canon_date is None:
                    raise ValueError(
                        "You must specify the canon_date property if you want to use a merged result query"
                    )
                self.explore_metrics[explore_name].append(field)

        for merged_metric in self.merged_metrics:
            for ref_field in merged_metric.referenced_fields(merged_metric.sql, ignore_explore=True):
                if isinstance(ref_field, str):
                    raise ValueError(f"Unable to find the field {ref_field} in the project")
                if ref_field.view.explore is None:
                    explore_name = merged_metric.view.explore.name
                else:
                    explore_name = ref_field.view.explore.name
                self.explore_metrics[explore_name].append(ref_field)

        for explore_name in self.explore_metrics.keys():
            self.explore_metrics[explore_name] = list(set(self.explore_metrics[explore_name]))

        dimension_mapping = defaultdict(list)
        for explore_name, field_set in self.explore_metrics.items():
            if len({f.canon_date for f in field_set}) > 1:
                raise NotImplementedError(
                    "Zenlytic does not currently support different canon_date "
                    "values for metrics in the same query in the same explore"
                )
            canon_date = field_set[0].canon_date
            for other_explore_name, other_field_set in self.explore_metrics.items():
                if other_explore_name != explore_name:
                    other_canon_date = other_field_set[0].canon_date
                    other_view_name = other_field_set[0].view.name
                    canon_date_data = {
                        "field": f"{other_view_name}.{other_canon_date}",
                        "explore_name": other_explore_name,
                    }
                    dimension_mapping[canon_date].append(canon_date_data)

        self.explore_dimensions = defaultdict(list)
        for dimension in self.dimensions:
            explore_name = self.project.get_explore_from_field(dimension)
            if explore_name not in self.explore_metrics:
                raise ValueError(
                    f"Could not find a metric in {self.metrics} that references the explore {explore_name}"
                )
            field = self.project.get_field(dimension, explore_name=explore_name)
            dimension_group = field.dimension_group
            self.explore_dimensions[explore_name].append(field)
            for mapping_info in dimension_mapping[field.name]:
                key = f"{mapping_info['field']}_{dimension_group}"
                field = self.project.get_field(key, explore_name=mapping_info["explore_name"])
                self.explore_dimensions[mapping_info["explore_name"]].append(field)

        self.explore_where = defaultdict(list)
        for where in self.where:
            explore_name = self.project.get_explore_from_field(where["field"])
            if explore_name not in self.explore_metrics:
                raise ValueError(
                    f"Could not find a metric in {self.metrics} that references the explore {explore_name}"
                )
            field = self.project.get_field(where["field"], explore_name=explore_name)
            dimension_group = field.dimension_group
            self.explore_where[explore_name].append(where)
            for mapping_info in dimension_mapping[field.name]:
                key = f"{mapping_info['field']}_{dimension_group}"
                field = self.project.get_field(key, explore_name=mapping_info["explore_name"])
                mapped_where = deepcopy(where)
                mapped_where["field"] = field.id(view_only=True)
                self.explore_where[mapping_info["explore_name"]].append(mapped_where)



import os
import shutil
from glob import glob

import git
import requests

from metrics_layer.core import utils

BASE_PATH = os.path.dirname(__file__)


class BaseRepo:
    def get_repo_type(self):
        if self.repo_type:
            return self.repo_type

        looker_files = list(self.search(pattern="*.model.*"))
        looker_files += list(self.search(pattern="*.view.*"))
        n_looker_files = len(looker_files)

        yaml_files = list(self.search(pattern="*.yml"))
        yaml_files += list(self.search(pattern="*.yaml"))
        n_yaml_files = len(yaml_files)

        # TODO Need to decide if we will support this
        # dbt_files = list(self.search(pattern="dbt_project.yml"))
        # dbt_files += list(self.search(pattern="dbt_project.yml"))
        # n_dbt_files = len(dbt_files)

        # if n_dbt_files > 0:
        #     return "dbt"
        if n_looker_files > n_yaml_files:
            return "lookml"
        return "metrics_layer"

    def delete(self):
        raise NotImplementedError()

    def search(self):
        raise NotImplementedError()

    def fetch(self):
        raise NotImplementedError()


class LocalRepo(BaseRepo):
    def __init__(
        self, repo_path: str, repo_type: str = None, dbt_path: str = None, warehouse_type: str = None
    ) -> None:
        self.repo_path = repo_path
        self.dbt_path = dbt_path
        if isinstance(dbt_path, str) and dbt_path[-1] != "/":
            self.dbt_path += "/"
        self.repo_type = repo_type
        self.warehouse_type = warehouse_type
        self.folder = f"{os.path.join(os.getcwd(), self.repo_path)}/"
        self.branch_options = []

    def search(self, pattern: str):
        """Example arg: pattern='*.model.*'"""
        return glob(f"{self.folder}**/{pattern}", recursive=True)

    def fetch(self):
        pass

    def delete(self):
        pass


class GithubRepo(BaseRepo):
    def __init__(self, repo_url: str, branch: str, repo_type: str = None, warehouse_type: str = None) -> None:
        self.repo_url = repo_url
        self.repo_type = repo_type
        self.warehouse_type = warehouse_type
        self.repo_name = utils.generate_uuid()
        self.repo_destination = os.path.join(BASE_PATH, self.repo_name)
        self.folder = f"{self.repo_destination}/"
        self.dbt_path = self.folder
        self.branch = branch
        self.branch_options = []

    def search(self, pattern: str):
        """Example arg: pattern='*.model.*'"""
        return glob(f"{self.folder}**/{pattern}", recursive=True)

    def fetch(self):
        self.fetch_github_repo(self.repo_url, self.repo_destination, self.branch)

        try:
            dynamic_branch_options = []
            g = git.cmd.Git()
            raw = g.ls_remote(self.repo_url)
            for raw_branch_ref in raw.split("\n"):
                if "/heads/" in raw_branch_ref:
                    clean_branch_ref = raw_branch_ref.split("/heads/")[-1]
                    dynamic_branch_options.append(clean_branch_ref)
            self.branch_options = dynamic_branch_options
        except Exception as e:
            print(f"Exception getting branch options: {e}")
            self.branch_options = [self.branch]

    def delete(self, folder: str = None):
        if folder is None:
            folder = self.folder

        if os.path.exists(folder) and os.path.isdir(folder):
            shutil.rmtree(folder)

    @staticmethod
    def fetch_github_repo(repo_url: str, repo_destination: str, branch: str):
        if os.path.exists(repo_destination) and os.path.isdir(repo_destination):
            shutil.rmtree(repo_destination)
        repo = git.Repo.clone_from(repo_url, to_path=repo_destination, branch=branch, depth=1)
        return repo


class LookerGithubRepo(BaseRepo):
    def __init__(
        self, looker_url: str, client_id: str, client_secret: str, project_name: str, repo_type: str = None
    ):
        self.looker_url = looker_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.project_name = project_name
        self.repo_type = repo_type
        self.repo_url, self.branch = self.get_looker_github_info()
        self.repo = GithubRepo(self.repo_url, self.branch)
        self.dbt_path = self.repo.folder
        self.branch_options = []

    def search(self, pattern: str):
        """Example arg: pattern='*.model.*'"""
        return self.repo.search(pattern=pattern)

    def fetch(self) -> None:
        self.repo.fetch()
        self.branch_options = self.repo.branch_options

    def delete(self, folder: str = None) -> None:
        self.repo.delete(folder=folder)

    def get_looker_github_info(self):
        projects = self.get_looker_projects()
        project = next((p for p in projects if p["name"] == self.project_name))
        return project["git_remote_url"], project["git_production_branch_name"]

    def get_looker_projects(self):
        token = self.get_looker_oauth_token(self.looker_url, self.client_id, self.client_secret)
        headers = {"Authorization": f"token {token}"}
        response = requests.get(f"{self.looker_url}/api/3.1/projects", headers=headers)
        return response.json()

    @staticmethod
    def get_looker_oauth_token(looker_url, client_id, client_secret):
        data = {"client_id": client_id, "client_secret": client_secret}
        response = requests.post(f"{looker_url}/api/3.1/login", data=data)
        if response.status_code == 403:
            raise ValueError("Looker credentials not valid, please check your credentials")
        return response.json()["access_token"]



import json
import os
from argparse import ArgumentError
from copy import deepcopy

from metrics_layer.core.model.definitions import Definitions


class MetricsLayerConnectionError(Exception):
    pass


class ConnectionType:
    snowflake = Definitions.snowflake
    bigquery = Definitions.bigquery
    redshift = Definitions.redshift


class BaseConnection:
    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name}>"


class SnowflakeConnection(BaseConnection):
    def __init__(
        self,
        name: str,
        account: str,
        password: str,
        role: str = None,
        username: str = None,
        user: str = None,
        warehouse: str = None,
        database: str = None,
        schema: str = None,
        **kwargs,
    ) -> None:
        self.type = ConnectionType.snowflake
        self.name = name
        self.account = account
        if user and username:
            raise ArgumentError(
                "Received arguments for both user and username, "
                "please send only one argument for the Snowflake user"
            )
        elif username:
            self.username = username
        elif user:
            self.username = user
        else:
            raise ArgumentError("Received no argument for the Snowflake user, pass either user or username")
        self.password = password
        self.role = role
        self.warehouse = warehouse
        self.database = database
        self.schema = schema

    def to_dict(self):
        """Dict for use with the snowflake connector"""
        base = {
            "name": self.name,
            "user": self.username,
            "password": self.password,
            "account": self.account,
            "type": self.type,
        }
        if self.warehouse:
            base["warehouse"] = self.warehouse
        if self.database:
            base["database"] = self.database
        if self.schema:
            base["schema"] = self.schema
        if self.role:
            base["role"] = self.role
        return base

    def printable_attributes(self):
        attributes = deepcopy(self.to_dict())
        attributes.pop("password")
        attributes["name"] = self.name
        sort_order = ["name", "type", "account", "user", "database", "schema", "warehouse", "role"]
        return {key: attributes.get(key) for key in sort_order if attributes.get(key) is not None}


class RedshiftConnection(BaseConnection):
    def __init__(
        self,
        name: str,
        host: str,
        username: str,
        password: str,
        port: int = 5439,
        database: str = None,
        schema: str = None,
        **kwargs,
    ) -> None:
        self.type = ConnectionType.redshift
        self.name = name
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.database = database
        self.schema = schema

    def to_dict(self):
        base = {
            "name": self.name,
            "host": self.host,
            "port": self.port,
            "username": self.username,
            "password": self.password,
            "type": self.type,
        }
        if self.database:
            base["database"] = self.database
        if self.schema:
            base["schema"] = self.schema
        return base

    def printable_attributes(self):
        attributes = deepcopy(self.to_dict())
        attributes.pop("password")
        attributes["name"] = self.name
        sort_order = ["name", "type", "host", "port", "username", "database", "schema"]
        return {key: attributes.get(key) for key in sort_order if attributes.get(key) is not None}


class BigQueryConnection(BaseConnection):
    def __init__(self, name: str, credentials: str = None, keyfile: str = None, **kwargs) -> None:
        self.type = ConnectionType.bigquery
        self.name = name
        creds_to_use = credentials if credentials else keyfile
        self.credentials = self._convert_json_if_needed(creds_to_use, kwargs)
        self.project_id = self.credentials["project_id"]

    def to_dict(self):
        """Dict for use with the BigQuery connector"""
        return {
            "name": self.name,
            "credentials": self.credentials,
            "project_id": self.project_id,
            "type": self.type,
        }

    def printable_attributes(self):
        attributes = deepcopy(self.to_dict())
        attributes.pop("credentials")
        attributes["name"] = self.name
        sort_order = ["name", "type", "project_id"]
        return {key: attributes.get(key) for key in sort_order if attributes.get(key) is not None}

    @staticmethod
    def _convert_json_if_needed(creds: dict, kwargs: dict):
        if isinstance(creds, dict):
            return deepcopy(creds)
        elif isinstance(creds, str):
            try:
                return json.loads(creds)
            except json.JSONDecodeError:
                # This means it's a file path not a JSON string
                if os.path.isabs(creds):
                    path = creds
                else:
                    path = os.path.join(kwargs["directory"], creds)

                with open(path, "r") as f:
                    return json.load(f)
        else:
            raise TypeError(f"BigQuery credentials json had wrong type: {type(creds)} for value {creds}")



import random
import string
import uuid


def generate_uuid(db_safe=False):
    if db_safe:
        return generate_random_password(40)
    return str(uuid.uuid4())


def generate_random_password(length):
    # Random string with the combination of lower and upper case
    letters = string.ascii_letters
    result_str = "".join(random.choice(letters) for i in range(length))
    return result_str


def merge_nested_dict(base: dict, additional: dict):
    for key, val in additional.items():
        if isinstance(val, dict):
            if key in base and isinstance(base[key], dict):
                merge_nested_dict(base[key], additional[key])

        elif isinstance(val, list):
            if key in base and isinstance(base[key], list):
                additional[key] = merge_list(base[key], additional[key])

        else:
            if key in base:
                additional[key] = base[key]

    for key, val in base.items():
        if key not in additional:
            additional[key] = val

    return additional


def merge_list(base_list: list, additional_list: list):
    final_list = []
    added = []
    for item in base_list:
        if not isinstance(item, dict):
            final_list.append(item)
            continue

        additional_item = next((i for i in additional_list if i["name"] == item["name"]), None)
        if additional_item:
            added.append(item["name"])
            final_list.append({**additional_item, **item})
        else:
            final_list.append(item)
    for item in additional_list:
        if isinstance(item, dict) and item["name"] not in added:
            final_list.append(item)
    return final_list
